/*  FUNNEL LIBRARY
**  
**  Hashtable -- support for hashtables in Funnel
**  
**  09/08/2K, Matthias Zenger
*/

newtype HashtableEntry[a, b] = {
    def hash: int
    def key: a
    def value: b
    def toString: string
}

newtype Hashtable[a, b] = {
    def size: int
    def isEmpty: boolean
    def keys: List[a]
    def elements: List[b]
    def contains(vl: b): boolean
    def containsKey(key: a): boolean
    def apply(key: a): b
    def get(key: a): b
    def put(key: a, vl: b): ()
    def remove(key: a): ()
    def clear: ()
    def toString: string
}

newtype HashtableModule = {
    def create[a, b]: Hashtable[a, b]
    def createWith[a, b](buckets: int, load: float): Hashtable[a, b]
}

let Hashtable: HashtableModule = {
    def create[a, b] = createWith[a, b](59, 0.75f)
    def createWith[a, b](buckets: int, load: float) = {
        def size & state(count: int, t: int, s) = count & state(count, t, s)
        def isEmpty & state(count: int, t, s) = (count == 0) & state(count, t, s)
        def keys & state(c, t, table: Array[List[HashtableEntry[a, b]]]) = {
            val res = table.toList.foldl
                (y: List[a], x| y.append(x.map(z| z.key)))(List.Nil)
            res & state(c, t, table)
        }
        def elements & state(c, t, table) = {
            val res = table.toList.foldl
                (y: List[b], x| y.append(x.map(z| z.value)))(List.Nil)
            res & state(c, t, table)
        }
        def contains vl & state(c, t, table) = {
            val res = table.toList.foldl
              (found: boolean, xs: List[HashtableEntry[a, b]]|
                if (found) true else !xs.filter(x| x.value == vl).isEmpty) false
            res & state(c, t, table)
        }
        def containsKey key & state(c, t, table) = {
            val res = !table(key.hashCode % table.length).
                            filter(x| x.key == key).isEmpty
            res & state(c, t, table)
        }
        def apply key & state(c, t, table) = {
            val res = table(key.hashCode % table.length).
                            find(x| x.key == key).value
            res & state(c, t, table)
        }
        def put(key, value) = (remove key; put'(key, value))
        def put'(key: a, value: b): () & state(c, t, table) = {
            val hash = key.hashCode % table.length
            def toString = key + " => " + value
            table(hash) := List.Cons(current: HashtableEntry[a, b], table(hash))
            rehash & state(c + 1, t, table)
        }
        def remove key & state(c, t, table) = {
            val hash = key.hashCode % table.length
            var c': int := c
            table(hash) := table(hash).filter(x|
                if (x.key == key) (c' := c' - 1; false) else true)
            () & state(c', t, table)
        }
        def clear & state(c, t, table) = {
            table.updateAll(x| List.Nil)
            () & state(0, t, table)
        }
        def rehash & state(c, t, table) = {
            () & state(c, t, table)
        }
        def toString & state(c, t, table) = {
            val res = List.concat(table.toList).toString
            res & state(c, t, table)
        }
        val get = apply
        val table = Array.create[List[HashtableEntry[a, b]]](buckets)
        val x = table.updateAll(x| List.Nil)
        current & state(0, (buckets * load).toInt, table)
    }
}
