/*  FUNNEL LIBRARY
**  
**  List -- support for lists in Funnel
**  
**  03/08/2K, Matthias Zenger
*/

newtype List[a] = {
    // strict list functions
    def isEmpty: boolean
    def head: a
    def tail: List[a]
    def first: a
    def second: a
    def third: a
    def next: List[a]
    def snoc(y: a): List[a]
    def length: int
    def last: a
    def apply(n: int): a
    def update(n: int)(x: a): List[a]
    def get(n: int): a
    def take(n: int): List[a]
    def drop(n: int): List[a]
    def takeWhile(p: a -> boolean): List[a]
    def dropWhile(p: a -> boolean): List[a]
    def append(ys: List[a]): List[a]
    def reverse: List[a]
    def iterate(f: a -> ()): ()
    def map[b](f: a -> b): List[b]
    def foldl[b](f: (b, a) -> b)(y: b): b
    def foldr[b](f: (a, b) -> b)(y: b): b
    def zip[b](ys: List[b]): List[(a, b)]
    def zipWith[b, c](f: (a, b) -> c)(ys: List[b]): List[c]
    def filter(p: a -> boolean): List[a]
    def find(p: a -> boolean): a
    def sort: List[a]
    def sortWith(less: (a, a) -> boolean): List[a]
    def toString: string
    def ::(x: a): List[a]
    def ++(ys: List[a]): List[a]
    def <-[b](f: a -> b): List[b]
    def match[b](v: ListVisitor[a, b]): b
    // lazy list functions
    def appendL(ys: List[a]): List[a]
    def mapL[b](f: a -> b): List[b]
    def zipL[b](ys: List[b]): List[(a, b)]
    def zipWithL[b, c](f: (a, b) -> c)(ys: List[b]): List[c]
    def filterL(p: a -> boolean): List[a]
    def <~[b](f: a -> List[b]): List[b]
}

newtype ListVisitor[a, +b] = {
    def Nil: b
    def Cons(x: a, xs: List[a]): b
}

type ListModule = {
  /** constructors
   */
    def Nil[a]: List[a]
    def Cons[a](x: a, xs: List[a]): List[a]
    
    def make[a](x: a): List[a]
    def make2[a](x: a, y: a): List[a]
    def make3[a](x: a, y: a, z: a): List[a]
    def make4[a](x: a, b: a, c: a, d: a): List[a]
    def make5[a](x: a, b: a, c: a, d: a, e: a): List[a]
    def make6[a](x: a, b: a, c: a, d: a, e: a, f: a): List[a]
    def make7[a](x: a, b: a, c: a, d: a, e: a, f: a, g: a): List[a]
    def make8[a](x: a, b: a, c: a, d: a, e: a, f: a, g: a, h: a): List[a]
    def fromString(s: string): List[char]
    def fromFile(f: string): List[char]
    def numbers(f: int): List[int]
    
  /** static list functions
   */
    def charsToString(xs: List[char]): string
    def concat[a](xss: List[List[a]]): List[a]
}


let List: ListModule = {
  /** list constructors
   */
    def Nil[t] = {
        // strict list functions
        def isEmpty = true
        def head = System.error "list empty"
        def tail = System.error "list empty"
        def first = head
        def second = head
        def third = head
        def next = Nil[t]
        def snoc y = List.Cons(y, Nil[t])
        def length = 0
        def last = System.error "no last element"
        def apply n = System.error "list empty"
        def update n x = System.error "list empty"
        def get n = System.error "list empty"
        def take n = Nil[t]
        def drop n = Nil[t]
        def takeWhile p = Nil
        def dropWhile p = Nil
        def append ys = ys
        def reverse = Nil
        def iterate f = ()
        def map[s] f = Nil
        def foldl[s] f y = y
        def foldr[s] f y = y
        def zip[s] ys = Nil
        def zipWith[s, s'] f ys = Nil
        def filter p = Nil
        def find p = null
        def sort = Nil
        def sortWith less = Nil
        def toString = "[]"
        def :: x = Cons(x, Nil[t])
        def ++ ys = ys
        def <- [s] f = Nil
        def match[s] v = v.Nil
        // lazy list functions
        def appendL ys = ys
        def mapL[s] f = Nil
        def zipL[s] ys = Nil[(t, s)]
        def zipWithL[s, s'] f ys = Nil
        def filterL p = Nil
        def <~ [s] f = Nil
    }
    def Cons[t](x, xs) = {
        // strict list functions
        def isEmpty = false
        def head = x
        def tail = xs
        def first = x
        def second = xs.first
        def third = xs.second
        def next = xs
        def snoc y = Cons(y, Cons(x, xs))
        def length = 1 + xs.length
        def last = if (xs.isEmpty) x else xs.last
        def apply n = if (n <= 0) x else xs.apply(n - 1)
        def update n y = if (n <= 0) Cons(y,xs) else Cons(x,xs.update(n - 1)y)
        def get n = if (n <= 0) x else xs.get(n - 1)
        def take n = if (n > 0) Cons(x, xs.take(n - 1)) else List.Nil
        def drop n = if (n > 0) xs.drop(n - 1) else Cons(x, xs)
        def takeWhile p = if (p(x)) Cons(x, xs.takeWhile p) else List.Nil
        def dropWhile p = if (p(x)) xs.dropWhile p else Cons(x, xs)
        def append ys = if (ys.isEmpty) Cons(x, xs) else Cons(x, xs.append(ys))
        def reverse = foldl[List[t]] (x, y| Cons(y, x)) (List.Nil)
        def iterate f = (f(x); xs.iterate f)
        def map[s] f = Cons(f(x), xs.map[s](f))
        def foldl[s] f y = xs.foldl f (f(y, x))
        def foldr[s] f y = f(x, (xs.foldr f y))
        def zip[s] ys = if (ys.isEmpty) Nil
                        else Cons((x, ys.head), xs.zip(ys.tail))
        def zipWith[s, s'] f ys = if (ys.isEmpty) Nil
                        else Cons(f(x, ys.head), xs.zipWith[s, s'] f (ys.tail))
        def filter p = if (p(x)) Cons(x, xs.filter p) else xs.filter p
        def find p = if (p(x)) x else xs.find p
        def sort = sortWith(x, y| x < y)
        def sortWith less = xs.filter(y| less(y, x)).sortWith(less).append
                        (List.Cons(x, xs.filter(y| !less(y, x)).sortWith less))
        def toString ="["+xs.foldl[string](x,y|x+", "+y.toString)(x.toString)+"]"
        def :: y = Cons(y, Cons(x, xs))
        def ++ ys = ys.append(Cons(x, xs))
        def <- = map
        def match[s] v = v.Cons(x, xs)
        // lazy list functions
        def appendL ys = Cons(x, (let zs = if (ys.isEmpty) xs
                                           else xs.appendL(ys); zs))
        def mapL[s] f = Cons(f(x), (let ys = xs.mapL[s](f); ys))
        def zipL[s] ys = if (ys.isEmpty) Nil
                      else Cons((x, ys.head), (let zs = xs.zipL(ys.tail); zs))
        def zipWithL[s, s'] f ys = if (ys.isEmpty) Nil else Cons(f(x, ys.head),
                                    (let zs = xs.zipWithL[s,s'] f(ys.tail);zs))
        def filterL p = (let ys = xs.filterL(p); if (p(x)) Cons(x, ys) else ys)
        def <~ [s] f = {
            f(x).append(val zs = xs.<~ f; zs)
        }
    }
    
  /** (strict) list constructors
   */
    def make[t](x) = Cons(x, Nil[t])
    def make2[t](x, y) = Cons(x, Cons(y, Nil[t]))
    def make3[t](x, y, z) = Cons(x, Cons(y, Cons(z, Nil[t])))
    def make4[t](a, b, c, d) = Cons(a, make3[t](b, c, d))
    def make5[t](a, b, c, d, e) = Cons(a, make4[t](b, c, d, e))
    def make6[t](a, b, c, d, e, f) = Cons(a, make5[t](b, c, d, e, f))
    def make7[t](a, b, c, d, e, f, g) = Cons(a, make6[t](b, c, d, e, f, g))
    def make8[t](a, b, c, d, e, f, g, h) = Cons(a,make7[t](b, c, d, e, f, g, h))
    def fromString str = {
        def from(str: string, xs: List[char], n: int): List[char] =
            if (n < 0) xs else from(str, Cons(str(n), xs), n - 1)
        from(str, Nil[char], str.length - 1)
    }
    
  /** lazy list constructors
   */
    def fromFile fname = {
        val in = Java.jconstructor (Java.java_lang_String)
                    (Java.jclass "java.io.FileInputStream") fname
        def read = {
            val c = Java.jmethod "read" () in
            if (c == -1) (Java.jmethod "close" () in; Nil)
            else Cons(c.toChar, (let cs = read; cs))
        }
        read
    }
    def numbers(from) = Cons(from, (let xs = numbers(from + 1); xs))
  
  /** projections
   */
    def charsToString(xs: List[char]): string = {
        if (xs.isEmpty)
            ""
        else
            xs.tail.foldl[string](s, y| s + y)("" + xs.head)
    }
    
  /** static list functions
   */
    def concat[t](xss: List[List[t]]): List[t] = {
        xss.foldr(xs: List[t], ys: List[t]| xs.append(ys))(List.Nil)
    }
}

def Nil[a]: List[a] = List.Nil
