/*  FUNNEL LIBRARY
**  
**  Mail -- an implementation of a small e-mail client written in
**          Funnel. Mails are sent via the SMTP protocol.
**          Mails are read from a POP server.
**  
**  28/06/2K, Matthias Zenger
*/


newtype MailResult[+a] = {
    def succeeded: boolean
    def message: string
    def result: a
    def <-[b](k: a -> MailResult[b]): MailResult[b]
}

newtype SMTPClient = {
    def messageTo(to: string, subject: string, message: string): MailResult[()]
}

newtype POPClient = {
    def count(password: string): int
    def message(password: string, msgnumber: int): string
}

newtype MailModule = {
    def sendFrom(server: string, port: int, from: string): SMTPClient
    def receiveFrom(server: string, port: int, user: string): POPClient
}

let Mail: MailModule = {

  /** import string operations
   */
    with String
    
  /** sending mail was successfull
   */
    def return[a](result: a): MailResult[a] = {
        val succeeded = true
        val message = "done"
        def <- [b] k = k result
    }
    
  /** there was an error detected
   */
    def throw[a](message: string): MailResult[a] = {
        def succeeded = false
        def result = System.error "no result"
        def <- [b] _ = throw[b](message)
    }
    
  /** create a client for an SMTP server at the given 'host', 'port'
   *  and for a special sender 'from'; this client offers a single
   *  method 'send' for sending mails; here's an example usage for
   *  sending mail as user matthias.zenger@epfl.ch to zenger@acm.org:
   *
   *  Mail.sendFrom("dimail.epfl.ch", 25, "matthias.zenger@epfl.ch")
   *      .messageTo("zenger@acm.org", "subject", "and this is the body")
   */
    def sendFrom(server, port, from) = {
        def messageTo(to, subject, message): MailResult[()] = {
            val EOL = "\r\n"
            val socket = Socket.create(server, port)
            val in = DataInputStream.create(socket.getInputStream)
            val out = PrintWriter.create(socket.getOutputStream)
            var res: string := in.readLine
            Predef.while (|startsWith res "220-") {|
                res := in.readLine
            }
            if (((indexOf res '-') == 3) || !(startsWith res "220")) {
                socket.close
                throw "protocol violation; 220 expected"
            } else {
                def send(message: string): MailResult[()] = {
                    out.writeString(message + EOL)
                    out.flush
                    return ()
                }
                def sendConf(message: string, confirmation: string): MailResult[()] = {
                    val res = send(message)
                    val answer = in.readLine
                    if (startsWith answer confirmation)
                        return ()
                    else {
                        socket.close
                        throw("host sent '" + answer + "', but " + confirmation + " expected")
                    }
                }
                val () <- sendConf("HELO " + InetAddress.getLocalHost.getHostName, "250")
                val () <- sendConf("MAIL FROM: " + from, "250")
                val () <- sendConf("RCPT TO: " + to, "250")
                val () <- sendConf("DATA", "354")
                val () <- send("From: " + from)
                val () <- send("To: " + to)
                val () <- send("Subject: " + subject)
                val () <- send("X-Mailer: Funnel Mail by Matthias Zenger")
                val () <- send("")
                val () <- send(message)
                val () <- sendConf(".", "250")
                socket.close
                return ()
            }
        }
    }

  /** create a client for a POP3 server at the given 'host', 'port'
   *  and for a special user 'user'; this client offers a couple of
   *  methods:
   *      count(password):
   *          returns the number of messages in the mailbox
   *      message(password, msgnumber):
   *          returns message number msgnumber as a string
   *
   *  usage:
   *    Mail.receiveFrom("pop.provider.xy", 110, "mzenger")
   *        .check("mypassword")
   */
    def receiveFrom(server, port, user) = {
        type POP3Interface = {
            def close: ()
            def send(message: string): MailResult[()]
            def sendConf(message: string): MailResult[string]
            def readLine: string
            def ready: boolean
        }
        def open: POP3Interface = {
            val EOL = "\r\n"
            val CONFIRMATION = "+OK "
            val socket = Socket.create(server, port)
            val in = DataInputStream.create(socket.getInputStream)
            val out = PrintWriter.create(socket.getOutputStream)
            def send(message) = {
                out.writeString(message + EOL)
                out.flush
                return ()
            }
            def sendConf(message) = {
                out.writeString(message + EOL)
                out.flush
                val answer = in.readLine
                if (startsWith answer CONFIRMATION)
                    return (substring answer 4 (length answer))
                else {
                    socket.close
                    throw ("host sent '" + answer + "'")
                }
            }
            def readLine = in.readLine
            def close = socket.close
            if (startsWith (in.readLine) CONFIRMATION) {
                val ready = true
            } else {
                socket.close
                val ready = false
            }
        }
        def firstNum(str: string, n: int, v: int): int = {
            if ((n >= (length str)) || ((charAt str n) < '0'))
                v
            else
                firstNum(str, n + 1, 10 * v + ((charAt str n) - '0'))
        }
        def check(password: string): MailResult[int] = {
            val server = open
            if (server.ready) {
                val res <- server.sendConf("USER " + user)
                val msg <- server.sendConf("PASS " + password)
                val res <- server.sendConf("STAT")
                val bye <- server.send("QUIT")
                server.close
                return (firstNum(res, 0, 0))
            } else
                throw ("could not open host")
        }
        def count(password) = {
            val res = check(password)
            if (res.succeeded) res.result else -1
        }
        def read(password: string, msgnumber: int): MailResult[string] = {
            val server = open
            val res <- server.sendConf("USER " + user)
            val msg <- server.sendConf("PASS " + password)
            val cmd <- server.sendConf("RETR " + msgnumber)
            var i: int := firstNum(cmd, 0, 0)
            var res: string := ""
            Predef.while (| i > 0) {|
                val line = server.readLine;
                i := i - (length line) - 2;
                if (i > 0)
                    res := res + line + "\n"
                else
                    ()
            }
            val bye = server.send("QUIT")
            server.close
            return res
        }
        def message(password, msgnumber) = {
            val res = read(password, msgnumber)
            if (res.succeeded) res.result else "could not retrieve mail"
        }
    }
}
