;;; funnel.el - Major mode for editing Funnel code.
;;; Michel Schinz, May 10, 2000.

;; A faire :
;; code d'indentation
;; menu
;; Imenu si possible
;; mettre dans la doc du mode et ici un lien sur une page WWW pour nouv. vers.
;; support pour les types
;; menu (et touches ?) pour lancer la compilation

(require 'easymenu)
(require 'inferior-funnel-mode)

(defconst funnel-mode-version "0.1.1")
(defconst funnel-bug-e-mail "Michel.Schinz@epfl.ch")
(defconst funnel-web-url "http://lampwww.epfl.ch/funnel/")

(defconst funnel-xemacsp (string-match "XEmacs" emacs-version)
  "True iff we are running in XEmacs")

;; Customization

(defgroup funnel
  nil
  "Mode to edit Funnel code."
  :group 'languages)

(defcustom funnel-indent-step 2
  "Indentation step."
  :type 'integer
  :group 'funnel)

(defcustom funnel-file-suffix "fn"
  "Suffix for Funnel source code files."
  :type 'string
  :group 'funnel)

(defcustom funnel-mode-hook nil
  "Hook run when Funnel mode starts."
  :type 'hook
  :group 'funnel)

;; Key mappings

(defvar funnel-mode-map nil
  "Keymap used in Funnel mode.")

(unless funnel-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map "\t" 'funnel-indent-line)
    (define-key map "\M-\t" 'funnel-outdent-line)
    (define-key map [backspace] 'backward-delete-char-untabify)
    (define-key map "}" 'funnel-electric-brace)

    (define-key map "\C-c\C-l" 'funnel-load-file)
    (define-key map "\C-c\C-r" 'funnel-eval-region)
    (define-key map "\C-c\C-b" 'funnel-eval-buffer)
    (define-key map "\C-c\C-z" 'funnel-switch-to-interpreter)

    (easy-menu-define funnel-menu-bar map "Funnel menu"
		      '("Funnel"
			["Run interpreter..."
			 run-funnel (not (funnel-interpreter-runningp))]
			["Reset interpreter"
			 funnel-reset-interpreter (funnel-interpreter-runningp)]
			["Quit interpreter"
			 funnel-quit-interpreter (funnel-interpreter-runningp)]
			["Load file in interpreter..."
			 funnel-load-file (funnel-interpreter-runningp)]
			["Switch to interpreter"
			 funnel-switch-to-interpreter (funnel-interpreter-runningp)]
			["Evaluate region"
			 funnel-eval-region (and (funnel-interpreter-runningp)
						 (funnel-region-definedp))]
			["Evaluate buffer"
			 funnel-eval-buffer (funnel-interpreter-runningp)]
			"---"
			["Browse Funnel Web site..."
			 funnel-browse-web-site t]
			["Customize..."
			 funnel-customize t]
			["Report bug..."
			 funnel-report-bug t]))
    (setq funnel-mode-map map)))

;; Syntax table

(defvar funnel-mode-syntax-table nil
  "Syntax table in use in Funnel mode buffers.")

(unless funnel-mode-syntax-table
  (let ((table (make-syntax-table)))
    ;; symbol components
    (modify-syntax-entry ?! "w" table)
    (modify-syntax-entry ?@ "w" table)
    (modify-syntax-entry ?# "w" table)
    (modify-syntax-entry ?$ "w" table)
    (modify-syntax-entry ?^ "w" table)
    (modify-syntax-entry ?~ "w" table)
    (modify-syntax-entry ?' "w" table)
    (modify-syntax-entry ?_ "w" table)

    ;; punctuation characters
    (modify-syntax-entry ?. "." table)

    ;; comments
    (if funnel-xemacsp
	(modify-syntax-entry ?/  ". 1456" table)
      (modify-syntax-entry ?/  ". 124b" table))
    (modify-syntax-entry ?*  ". 23"   table)
    (modify-syntax-entry ?\n "> b" table)
    (modify-syntax-entry ?\r "> b" table)
    (setq funnel-mode-syntax-table table)))

;; Indentation

(defvar funnel-cached-pps-to nil)
(make-variable-buffer-local 'funnel-cached-pps-to)
(defvar funnel-cached-pps-state nil)
(make-variable-buffer-local 'funnel-cached-pps-state)

(defun funnel-parse-partial-sexp ()
  "Equivalent to `parse-partial-sexp' from buffer start to point.
Tries to perform some cacheing to avoid rescanning the whole buffer often."
  (let ((from (point-min)) (to (point)))
    (unless (and funnel-cached-pps-to (= funnel-cached-pps-to to))
      (let* ((can-reuse-statep (and funnel-cached-pps-to (> to funnel-cached-pps-to)))
	     (real-from (if can-reuse-statep funnel-cached-pps-to from))
	     (real-state (if can-reuse-statep funnel-cached-pps-state nil)))
	(setq funnel-cached-pps-state
	      (parse-partial-sexp real-from to nil nil real-state))
	(setq funnel-cached-pps-to to)))
    funnel-cached-pps-state))

(defun funnel-clear-pps-cache (start end)
  "Clear cache used by `funnel-parse-partial-sexp'.
This function should be called whenever the buffer is changed. That is, it
should be part of `before-change-functions'."
  (when (and funnel-cached-pps-to
	     (< start funnel-cached-pps-to))
    (setq funnel-cached-pps-to nil)))

(defun funnel-last-definition-pos ()
  "Position of last (easily) identifiable top-level definition before point.
This can be used as a FROM argument for `parse-partial-sexp'"
  (point-min))				;[...] refine

(defun funnel-in-indented-blockp ()
  "True iff the point is in a block where identation matters."
  (let ((block-start (nth 1 (funnel-parse-partial-sexp))))
    (or (not block-start)
	(char-equal (char-after block-start) ?{))))

(defun funnel-at-indented-block-startp ()
  "True iff the point is before the first expression in an indented block.
See also `funnel-in-indented-blockp'."
  (save-excursion
    (forward-comment (- (buffer-size)))
    (char-equal (preceding-char) ?{)))

(defun funnel-in-indentationp ()
  "True iff the point is in the indentation of the current line."
  (<= (current-column) (current-indentation)))

(defun funnel-in-commentp ()
  "True iff the point is inside a comment."
  (nth 4 (funnel-parse-partial-sexp)))

(defun funnel-in-literalp ()
  "True iff the point is inside a string literal.

Works only if outside a comment (i.e. `funnel-in-commentp' must be
false for that function to work correctly)!"
  ;; Since strings cannot span across lines in Funnel, we do not need
  ;; to use the full power of `funnel-parse-partial-sexp'.
  (save-excursion
    (save-match-data
      (let ((pt (point)))
	(beginning-of-line)
	;; Move past all well-terminated strings
	(while (search-forward-regexp "\"\\([^\\\\\"]\\|\\\\\"\\)*\"" pt t))
	(search-forward-regexp "\"\\([^\\\\\"]\\|\\\\\"\\)*" pt t)))))

(defun funnel-previous-line-indentation ()
  "Return indentation of previous non-blank and non-comment line."
  (save-excursion
    (beginning-of-line)
    (forward-comment (- (buffer-size)))
    (current-indentation)))

(defun funnel-indent-line-to (column)
  "Indent current line to COLUMN and perhaps move point.
The point is moved iff it is currently in the indentation, in which
case it is brought to the end of that indentation. Otherwise it does
not move."
  (if (funnel-in-indentationp)
      (indent-line-to column)
    (save-excursion (indent-line-to column))))

(defun funnel-indent-rigidly ()
  "Indent like previous line or one step further right.
When at the beginning of the line, indent like the previous non-blank line.
Otherwise indent one step further right."
  (let ((new-indentation (if (bolp)
			     (max (+ (funnel-previous-line-indentation)
				     (if (funnel-at-indented-block-startp)
					 funnel-indent-step
				       0))
				  funnel-indent-step)
			   (+ funnel-indent-step (current-indentation)))))
    (funnel-indent-line-to new-indentation)))

(defun funnel-indent-from-syntax ()
  "Not implemented yet."
  (funnel-indent-rigidly))		;[...] implement

(defun funnel-outdent-rigidly ()
  "Indent just right of previous line or one step further left.
When at the beginning of the line, indent one step left of the previous
non-blank line. Otherwise, indent one step further left."
  (let ((new-indentation (max 0 (if (bolp)
				    (- (funnel-previous-line-indentation)
				       funnel-indent-step)
				  (- (current-indentation) funnel-indent-step)))))
    (funnel-indent-line-to new-indentation)))

(defun funnel-indent-line ()
  "Indent current line as smartly as possible."
  (interactive)
  (cond ((funnel-in-commentp) (insert ?\t))
	((funnel-in-indented-blockp) (funnel-indent-rigidly))
	(t (funnel-indent-from-syntax))))

(defun funnel-outdent-line ()
  "Outdent current line."
  (interactive)
  (funnel-outdent-rigidly))

(defun funnel-electric-brace (arg)
  "Insert a brace.

The current line is also reindented provided that:
* no numeric argument is supplied,
* the brace is on a line by itself (i.e. surrounded by spaces only), and
* the brace is not inside a comment or a string literal."
  (interactive "*P")
  (unless (or arg
	      (not (funnel-in-indentationp))
	      (not (looking-at "[ \t]*$"))
	      (funnel-in-commentp)
	      (funnel-in-literalp))
    (funnel-indent-line-to (if (funnel-at-indented-block-startp)
			       (funnel-previous-line-indentation)
			     (max 0 (- (funnel-previous-line-indentation)
				       funnel-indent-step)))))
  (self-insert-command (prefix-numeric-value arg)))

;; Font lock

(defun funnel-match-declaration-and-skip (limit)
  "Match a declaration element and put point after it."
  (when (looking-at "\\(\\<def\\>\\|\\<val\\>\\|\\<let\\>\\|[&,]\\)?\\s *\\(\\(\\sw\\|\\.\\)+\\)[(), \t]*")
    (goto-char (match-end 0))
    ;; Skip over right hand side to next definition, if there is one.
    ;[...] on devrait sauter les commentaires ici aussi
    (when (char-equal (following-char) ?=)
      (save-match-data
	(save-restriction
	  (narrow-to-region (point-min) limit)
	  (condition-case nil
	      (progn (while (and (not (eobp))
				 (not (looking-at "\\s *,")))
		       (forward-sexp))
		     (skip-chars-forward "\t "))
	    (error nil)))))
    t))

(defvar funnel-font-lock-keywords
  `(;; keywords
    ("\\<\\(current\\|def\\|else\\|i\\(f\\|mport\\)\\|let\\|primitive\\|type\\|va[lr]\\|with\\)\\>"
    0 font-lock-keyword-face keep nil)
  ;; constants
  ("\\<\\(nil\\|true\\|false\\)\\>"
   0 ,(if funnel-xemacsp 'font-lock-keyword-face 'font-lock-constant-face) nil nil)
  ;; type definitions
  ("\\<type\\s +\\(\\sw+\\)"
   1 'font-lock-type-face keep nil)
  ;; variable definitions (we use `font-lock-function-name-face' for
  ;; similarity with "val" definitions)
  ("\\<var\\s +\\(\\sw+\\)\\s *:="
   1 'font-lock-function-name-face keep nil)
  ;; function and value definitions
  ("\\<\\(def\\|val\\|let\\)\\>"
   (funnel-match-declaration-and-skip (goto-char (match-beginning 0))
				      (goto-char (match-end 0))
				      (2 (if (match-beginning 1)
					     font-lock-function-name-face
					   font-lock-variable-name-face))))))

;; Misc

(defun funnel-region-definedp ()
  "True iff the region is currently defined."
  (condition-case nil
      (progn (region-beginning) t)
    (error nil)))

(defun funnel-browse-web-site ()
  "Browse the Funnel home-page"
  (interactive)
  (require 'browse-url)
  (browse-url funnel-web-url))

;; Bug reporting

(defun funnel-report-bug ()
  "Report a bug to the author of the Funnel mode via e-mail.
The package used to edit and send the e-mail is the one selected
through `mail-user-agent'."
  (interactive)
  (require 'reporter)
  (let ((reporter-prompt-for-summary-p t))
    (reporter-submit-bug-report
     funnel-bug-e-mail
     (concat "Funnel mode v" funnel-mode-version)
     '(funnel-indent-step
       funnel-file-suffix
       funnel-mode-hook))))

;; Customization

(defun funnel-customize ()
  "Customize Funnel mode using the Customize package."
  (interactive)
  (customize-group 'funnel))

;; Mode

(defun funnel-mode ()
  "Major mode for editing Funnel code.

When started, run `funnel-mode-hook'.

\\{funnel-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (setq major-mode 'funnel-mode)
  (setq mode-name "Funnel")

  ;; Font lock
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(funnel-font-lock-keywords
			     nil nil nil))

  ;; Syntax table
  (set-syntax-table funnel-mode-syntax-table)

  ;; Paragraph separation
  (make-local-variable 'paragraph-start)
  (setq paragraph-start (concat "^\\s *$\\|" page-delimiter))
  (make-local-variable 'paragraph-separate)
  (setq paragraph-separate paragraph-start)
  (make-local-variable 'paragraph-ignore-fill-prefix)
  (setq paragraph-ignore-fill-prefix t)

  ;; Comment handling
  (make-local-variable 'comment-start)
  (setq comment-start "// ")
  (make-local-variable 'comment-end)
  (setq comment-end "")
  (make-local-variable 'comment-multi-line)
  (setq comment-multi-line nil)
  (make-local-variable 'comment-start-skip)
  (setq comment-start-skip "// *")

  ;; Misc
  (make-local-variable 'require-final-newline)
  (setq require-final-newline t)

  (make-local-hook 'before-change-functions)
  (add-hook 'before-change-functions
	    (function funnel-clear-pps-cache) nil t)

  ;; Keymap & menus
  (use-local-map funnel-mode-map)
  (easy-menu-add funnel-menu-bar)

  (run-hooks 'funnel-mode-hooks))

;; Automatic mode selection for funnel files, if not already done.
(let ((suffix-regexp (concat "\\." funnel-file-suffix "\\'")))
  (unless (assoc suffix-regexp auto-mode-alist)
    (setq auto-mode-alist (cons (cons suffix-regexp 'funnel-mode)
				auto-mode-alist))))

;; Make sure that Funnel files are not ignored by completion or by
;; dired (this is typically the case if the standard ".fn" extension
;; is used).
(let ((funnel-extension (concat "." funnel-file-suffix)))
  (setq completion-ignored-extensions
	(delete funnel-extension completion-ignored-extensions)))

(provide 'funnel-mode)
