;;; inferior-funnel-mode - Major mode to interact with the Funnel interpreter
;;; Michel Schinz, May 11, 2000

(require 'comint)

;; Customization

(defgroup inferior-funnel
  nil
  "Mode to interact with a funnel interpreter."
  :group 'funnel
  :tag "Inferior Funnel")

(defcustom funnel-default-interpreter "funny"
  "Name of the interpreter to use by default."
  :type 'string
  :group 'inferior-funnel)

(defcustom inferior-funnel-mode-hook nil
  "Hook run when inferior Funnel mode starts."
  :type 'hook
  :group 'inferior-funnel)

(defvar inferior-funnel-interpreter funnel-default-interpreter
  "The interpreter that `run-funnel' should run.
If you want to change the defaut value, don't change that variable
but customize `funnel-default-interpreter' instead.")

(defconst inferior-funnel-buffer-name "*inferior-funnel*")

(defconst inferior-funnel-prompt-regexp "^[>|] "
  "Regular expression matching the Funnel interpreter's prompt")

(defvar inferior-funnel-mode-map nil
  "Keymap used in inferior Funnel mode.")

;; Keymap

(unless inferior-funnel-mode-map
  (let ((map (copy-keymap comint-mode-map)))
    (setq inferior-funnel-mode-map map)))

;; Font locking

(defun inferior-funnel-font-lock-keywords ()
  (append funnel-font-lock-keywords
	  (list (list "^[>|] " 0 (if funnel-xemacsp
				     'font-lock-preprocessor-face
				   'font-lock-warning-face) t nil))))

;; Mode
(defun inferior-funnel-mode ()
  "Major mode for interacting with a Funnel interpreter.

\\{inferior-funnel-mode-map}"
  (interactive)

  (comint-mode)
  (setq comint-prompt-regexp inferior-funnel-prompt-regexp)

  (setq major-mode 'inferior-funnel-mode)
  (setq mode-name "Inferior Funnel")

  ;; Font lock
  (make-local-variable 'font-lock-defaults)
  (setq font-lock-defaults '(inferior-funnel-font-lock-keywords
			     nil nil nil))

  ;; Syntax table
  (set-syntax-table funnel-mode-syntax-table)

  (run-hooks 'inferior-funnel-mode-hooks))

(defun run-funnel (interpreter)
  "Run a Funnel interpreter in an Emacs buffer"
  (interactive (list (if (funnel-interpreter-runningp)
			 inferior-funnel-interpreter
		       (read-from-minibuffer "Funnel interpreter to run: "
					     inferior-funnel-interpreter))))
  (setq inferior-funnel-interpreter interpreter)
  (unless (funnel-interpreter-runningp)
    (set-buffer (make-comint "inferior-funnel" inferior-funnel-interpreter))
    (inferior-funnel-mode))
  (switch-to-buffer inferior-funnel-buffer-name))

(defun funnel-interpreter-runningp ()
  "True iff a Funnel interpreter is currently running in a buffer."
  (comint-check-proc inferior-funnel-buffer-name))

(defun funnel-send-string (string)
  "Send a string to the Funnel interpreter, with prompt off."
  (comint-send-string inferior-funnel-buffer-name ":toggleprompt\n")
  (comint-send-string inferior-funnel-buffer-name string)
  (comint-send-string inferior-funnel-buffer-name "\n:toggleprompt\n"))

(defun funnel-eval-region (start end)
  "Send current region to Funnel interpreter."
  (interactive "r")
  (comint-send-string inferior-funnel-buffer-name ":toggleprompt\n")
  ;; send lines one after the other, putting a backslash after each
  ;; one ([...] for long regions, it might be faster to use a
  ;; temporary file).
  (save-excursion
    (save-restriction
      (narrow-to-region start end)
      (goto-char (point-min))
      (while (not (eobp))
	(let ((bol-pos (point)))
	  (end-of-line)
	  (comint-send-region inferior-funnel-buffer-name bol-pos (point))
	  (comint-send-string inferior-funnel-buffer-name "\\\n")
	  (beginning-of-line 2)))))
  (comint-send-string inferior-funnel-buffer-name "\n:toggleprompt\n"))

(defun funnel-eval-buffer ()
  "Send whole buffer to Funnel interpreter."
  (interactive)
  (funnel-eval-region (point-min) (point-max)))

(defun funnel-switch-to-interpreter ()
  "Switch to buffer containing the interpreter"
  (interactive)
  (switch-to-buffer inferior-funnel-buffer-name))

(defvar funnel-prev-l/c-dir/file nil
  "Caches the last (directory . file) pair.
Caches the last pair used in the last funnel-load-file.
Used for determining the default in the next one.")

(defun funnel-load-file (file-name &optional prefix)
  "Load a file in the Funnel interpreter."
  (interactive (comint-get-source "Load Funnel file: " funnel-prev-l/c-dir/file
				  '(funnel-mode) t))
  (comint-check-source file-name)
  (setq funnel-prev-l/c-dir/file (cons (file-name-directory    file-name)
				       (file-name-nondirectory file-name)))
  (funnel-send-string (concat ":l " file-name "\n")))

(defun funnel-reset-interpreter ()
  "Reset Funnel interpreter."
  (interactive)
  (funnel-send-string "\n:r\n"))

(defun funnel-quit-interpreter ()
  "Quit Funnel interpreter."
  (interactive)
  (funnel-send-string "\n:q\n"))

(provide 'inferior-funnel-mode)
